__title__ = "flippers"
__summary__ = "A Python library for weak supervision."
__url__ = "https://github.com/liamtoran/flippers"
__version__ = "0.0.1-beta4"
__author__ = "Liam Toran"
__author_email__ = "liam.toran@gmail.com"
__license__ = "Apache-2.0"
